/**
 * Copyright (c) 2023 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#pragma once
#include "dusk.h"

#ifndef ASSERTIONS_FAKED
  /**
   * Assert a given value to be true.
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param x Value to assert as true.
   * @param message Message to throw against assertion failure.
   */
  void assertTrueImpl(
    const char *file,
    const int32_t line,
    const bool_t x,
    const char *message
  );

  /**
   * Asserts a given statement to be false.
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param x Value to assert as false.
   * @param message Message to throw against assertion failure.
   */
  void assertFalseImpl(
    const char *file,
    const int32_t line,
    const bool_t x,
    const char *message
  );

  /**
   * Asserts that a given line of code is unreachable. Essentially a forced
   * assertion failure, good for "edge cases"
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param message Message to throw against assertion failure.
   */
  void assertUnreachableImpl(
    const char *file,
    const int32_t line,
    const char *message
  );

  /**
   * Assert a given pointer to not point to a null pointer.
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param pointer Pointer to assert is not a null pointer.
   * @param message Message to throw against assertion failure.
   */
  void assertNotNullImpl(
    const char *file,
    const int32_t line,
    const void *pointer,
    const char *message
  );

  /**
   * Asserts a given pointer to be a nullptr.
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param pointer Pointer to assert is nullptr.
   * @param message Message to throw against assertion failure.
   */
  void assertNullImpl(
    const char *file,
    const int32_t line,
    const void *pointer,
    const char *message
  );

  /**
   * Asserts a function as being deprecated.
   * 
   * @param file File that the assertion is being made from.
   * @param line Line that the assertion is being made from.
   * @param message Message to throw against assertion failure.
   */
  void assertDeprecatedImpl(
    const char *file,
    const int32_t line,
    const char *message
  );

  void assertMemoryRangeMatchesImpl(
    const char *file,
    const int32_t line,
    const void *start,
    const void *end,
    const size_t size,
    const char *message
  );

  #define assertTrue(x, message) \
    assertTrueImpl(__FILE__, __LINE__, x, message)

  #define assertFalse(x, message) \
    assertFalseImpl(__FILE__, __LINE__, x, message)

  #define assertUnreachable(message) \
    assertUnreachableImpl(__FILE__, __LINE__, message)

  #define assertNotNull(pointer, message) \
    assertNotNullImpl(__FILE__, __LINE__, pointer, message)

  #define assertNull(pointer, message) \
    assertNullImpl(__FILE__, __LINE__, pointer, message)

  #define assertDeprecated(message) \
    assertDeprecatedImpl(__FILE__, __LINE__, message)

  #define assertStrLenMax(str, len, message) \
    assertTrue(strlen(str) < len, message)

  #define assertStrLenMin(str, len, message) \
    assertTrue(strlen(str) >= len, message)

#else
  // If assertions are faked, we define the macros to do nothing.
  #define assertTrue(x, message) ((void)0)
  #define assertFalse(x, message) ((void)0)
  #define assertUnreachable(message) ((void)0)
  #define assertNotNull(pointer, message) ((void)0)
  #define assertNull(pointer, message) ((void)0)
  #define assertDeprecated(message) ((void)0)
  #define assertStrLenMax(str, len, message) ((void)0)
  #define assertStrLenMin(str, len, message) ((void)0)

#endif